﻿using System;
using System.Globalization;
using System.ServiceProcess;
using System.Threading;
using System.Timers;
using System.Diagnostics;



namespace ShardsOnlineWatchdog
{
    class Program
    {
        /// *** GLOBALS ***
        /// 
        /// <summary>
        /// Holds the last timestamp recieved from the text file
        /// </summary>
        private static DateTime m_LastCheck = DateTime.MinValue;
        /// <summary>
        /// Holds the last time we made a server start attempt
        /// </summary>
        private static DateTime m_LastRestart = DateTime.MinValue;
        /// <summary>
        /// Time to wait after a restart attempt to try another restart 
        /// </summary>
        private static int      m_Timeout     = Properties.Settings.Default.TimeOut;
        /// <summary>
        /// Heartbeat file written by the server heartbeat script
        /// </summary>
        private static string   m_fileName    = Properties.Settings.Default.ServerRoot +"\\"+ Properties.Settings.Default.HeartbeatFile;
        /// <summary>
        /// Name under which the Shard Server runs as Service
        /// </summary>
        private static string   m_ServiceName = Properties.Settings.Default.ServiceName;



        /// <summary>
        /// Checks service status after a timeout and restarts the Shard Service 
        /// </summary>
        public static void RestartShard()
        {
            ServiceController myservice = new ServiceController(m_ServiceName);
            /// Decide depending on service status
            switch (myservice.Status)
            {
                case ServiceControllerStatus.Running:
                    /// service is running but we got a timeout alarm ... nasty situation eh?
                    /// case 1: The server is frozen
                    /// case 2: The server has just restarted and not yet written a new Heartbeat file
                    if (((DateTime.UtcNow-m_LastRestart).Ticks/10000000 > m_Timeout) && (m_LastRestart != DateTime.MinValue))
                        /// means we will see this as a server freeze
                    {
                        Console.WriteLine("Shard Service is in Running state, but failed to write heartbeat file in due time ("+Properties.Settings.Default.TimeOut+" seconds). Service is hanging most likely. Trying to force restart ...");
                        StopShardService(myservice);
                    }
                    else
                        /// means we will see this as a pending startup
                    {
                        Console.WriteLine("Shard Service is in Running state, and failed to write heartbeat file. Timeout conditions (" + Properties.Settings.Default.TimeOut + " seconds) for a forceful restart not yet met. We will wait ...");
                    }


                    return;
                case ServiceControllerStatus.Stopped:
                    //Service is down - restarting
                    Console.WriteLine("Shard Service " + myservice.DisplayName + " is in Stopped state. Trying to restart ...");
                    StartShardService(myservice);
                    return;
                case ServiceControllerStatus.Paused:
                    Console.WriteLine("Weird! Shard Service "+myservice.DisplayName+" is in Paused state .... ");
                    //That should never happen
                    return;
                case ServiceControllerStatus.StopPending:
                    Console.WriteLine("Shard Service " + myservice.DisplayName + " is in Stop Pending state. Lets wait a little ... ");
                    //We will just wait another interval until it is stopped
                    return;
                case ServiceControllerStatus.StartPending:
                    Console.WriteLine("Shard Service " + myservice.DisplayName + " is in Start Pending state. Lets wait a little ... ");
                    ///Nothing to do here - all looks good
                    return;
                default:
                    /// This is really weird ....
                    Console.WriteLine("Now this is really weird!!! Shard Service " + myservice.DisplayName + " is in "+myservice.Status+" state. Lets wait a little ... ");
                    return;
            }

        }

        /// <summary>
        /// Starting the Shard Service
        /// </summary>
        /// <param name="myservice">Service to Start</param>
        public static void StartShardService(ServiceController myservice)
{
            /// restart attempt timestamp
            m_LastRestart = DateTime.UtcNow;
            try
            {
                TimeSpan timeout = TimeSpan.FromMilliseconds(10000);
                Console.WriteLine("Trying to start Service");
                myservice.Start() ;
                myservice.WaitForStatus(ServiceControllerStatus.Running, timeout);
            }
            catch (Exception ex)
            {   /// The restart failed - now forcefully restarting by asking an external program to do it
                Console.WriteLine("Friendly Start failed: "+ex.Message);
            }
        }

        /// <summary>
        /// Stopping the Shard Service
        /// </summary>
        /// <param name="myservice">Service to stop</param>
        public static void StopShardService(ServiceController myservice)
        {
            try
            {
                myservice.Stop();
            }
            catch (Exception ex2)
            {
                Console.WriteLine("An error occoured while stopping the Service: " + ex2.Message);
                return;
            }
        }



        /// <summary>
        /// Convert Unix Time Stamp to DateTime
        /// </summary>
        /// <param name="unixTimeStamp"></param>
        /// <returns></returns>
        public static DateTime UnixTimeStampToDateTime(double unixTimeStamp)
        {
            // Unix timestamp is seconds past epoch
            System.DateTime dtDateTime = new DateTime(1970, 1, 1, 0, 0, 0, 0, System.DateTimeKind.Utc);
            dtDateTime = dtDateTime.AddSeconds(unixTimeStamp).ToLocalTime();
            return dtDateTime;
        }

        /// <summary>
        /// Convert DateTime to Timestamp
        /// </summary>
        /// <param name="dateTime"></param>
        /// <returns></returns>
        public static double DateTimeToUnixTimestamp(DateTime dateTime)
        {
            return (dateTime - new DateTime(1970, 1, 1).ToLocalTime()).TotalSeconds;
        }

        /// <summary>
        /// This Eventhandler checks the file for the right date
        /// </summary>
        /// <param name="source"></param>
        /// <param name="e"></param>
        static private void myEvent(object source, ElapsedEventArgs e)
        {
            ///Console.Write("DebugMessage: Elapsed Event - " + DateTime.Now.ToString()+ "\n");
            string line;
            ///Console.Write("DebugMessage: Elapsed Event - " + m_fileName + "\n");
            System.IO.StreamReader file = new System.IO.StreamReader(m_fileName);
            line = file.ReadLine();
            file.Close();
            

            try
            {
                ///DateTime currentStamp = DateTime.Parse(line);
                DateTime currentStamp = DateTime.ParseExact(line, "M'/'d'/'yyyy h':'mm':'ss tt",CultureInfo.InvariantCulture);
                /// TimeSpan delta = DateTime.UtcNow - currentStamp;
                TimeSpan delta = DateTime.UtcNow - currentStamp;
                Console.Write("Checking Heartbeat: (HB " + line + ", Delta=" + delta.ToString("%d'd 'hh':'mm':'ss") + ") checked at " + DateTime.UtcNow.ToLongTimeString() + "\r" );

                /// if not yet initialized ...
                if (m_LastCheck == DateTime.MinValue)
                {
                    m_LastCheck = currentStamp;
                    Console.WriteLine("Initialized on:" + currentStamp.ToLongTimeString());

                    /// if (delta.TotalSeconds >= m_Timeout)
                    /// {
                    ///Console.WriteLine("Timestamp is too old to initialize!");
                    ///System.Environment.Exit(1);
                    ///}
                }
                else
                {
                    ///Console.WriteLine("DebugMessage: Delta: " + delta.TotalSeconds.ToString() + " seconds");
                    if (delta.TotalSeconds >= Properties.Settings.Default.HeartbeatTolerance)
                    {
                     Console.WriteLine("Heartbeat Timeout: " + delta.TotalSeconds.ToString() + " seconds - restarting shard service ...");
                     RestartShard();
                    }
                }
            }
            catch (Exception ex)
            {
                  Console.WriteLine("Error parsing Datestring! "+"("+line+")"+ ex.Message+"\n\r");
            }
        }



        static int Main(string[] args)
        {
            var myTimer = new System.Timers.Timer();
            myTimer.Elapsed += new ElapsedEventHandler(myEvent);
            myTimer.Interval = Properties.Settings.Default.CheckIntervalSecs*1000;       // Set Time Interval
            myTimer.Start();

            Console.WriteLine("Watchdog Running..."
                + "\nStarting Time:   " + DateTime.UtcNow.ToLocalTime()
                + "\nService Name:    " + Properties.Settings.Default.ServiceName
                + "\nShard Root Dir:  " + Properties.Settings.Default.ServerRoot
                + "\nHeartbeat file:  " + Properties.Settings.Default.HeartbeatFile
                + "\nCheck Interval:  " + Properties.Settings.Default.CheckIntervalSecs + " seconds"
                + "\nLag Tolerance:   " + Properties.Settings.Default.HeartbeatTolerance
                + "\nRestart Timeout: " + Properties.Settings.Default.TimeOut + " seconds"
                );

            if (!System.IO.File.Exists(Properties.Settings.Default.ServerRoot+"\\"+Properties.Settings.Default.HeartbeatFile))
            {
                Console.WriteLine("Watchdog file not found!");
                System.Environment.Exit(1);
            }
            else
            {
                Console.WriteLine("External file found!");
                
            }

            while (Console.Read() != 'q')
            {
                Thread.Sleep(1000);
            }

            return 0;
        }
    }
}
