-- Character relation system (revision 3)

ORDER_PRIORITY = {"add", "multiply", "divide", "set", "setmin", "setmax"}

orderFunctions = {}
function orderFunctions.add(sympathy, mod)
	return mod
end

function orderFunctions.multiply(sympathy, mod)
	return sympathy * mod - sympathy
end

function orderFunctions.divide(sympathy, mod)
	return sympathy / mod - sympathy
end

function orderFunctions.set(sympathy, mod)
	return sympathy + mod
end

function orderFunctions.setmin(sympathy, mod)
	if sympathy < mod then 
		return mod - sympathy
	else 
		return 0
	end
end

function orderFunctions.setmax(sympathy, mod)
	if sympathy > mod then 
		return mod - sympathy
	else 
		return 0
	end
end

-- List of all traits this character has
Traits = {}

-- List of all traits this character cares about
Concerns = {}

-- Cached relation values
RelationCache = {}

-- Sets a trait
function SetTrait(name,level)
	Traits[name] = level
	Save()
end

-- Modifies a trait with given value
function ModifyTrait(name,mod)
	local old = Traits[name] or 0
	Traits[name] = old + mod
	Save()
end

-- Sets a trait to care about
function SetConcern(name,orders)
	Concerns[name] = orders
	Save()
end

-- Calculates (if needed) and returns a sympathy value
function GetSympathy(target)
	-- Cache check
	if RelationCache[target] ~= nil then
		if DateTime.UtcNow.Ticks - RelationCache[target].update < 180000000000 then
			return RelationCache[target].value
		else
			RelationCache[target] = nil
		end
	end

	local traits = target:GetObjVar("RelTraits")
	if traits == nil then return 0 end

	-- 
	local calcOrders = {}
	-- initial calculation
	if #traits < #Concerns then -- Optimization, select the shorter table for pairs()
		for name,level in pairs(traits) do
			if Concerns[name] ~= nil then
				calcOrders[name] = Concerns[name]
			end
		end
	else
		for name,usage in pairs(Concerns) do
			if traits[name] ~= nil then
				calcOrders[name] = usage
			end
		end
	end

	local sympathy = 0
	for priority,name in ipairs(ORDER_PRIORITY) do -- Loop throught order priorities
		local result = 0
		local func = orderFunctions[name]
		for trait,value in pairs(calcOrders) do -- Loop through traits for this order type
			local mod = value[name]
			if mod ~= nil then
				result = result + func(sympathy, value[name]) * traits[trait]
			end
		end

		sympathy = sympathy + result
	end

	CacheRelation(target, sympathy)
	return sympathy
end

-- Saves all data
function Save()
	this:SetObjVar("RelTraits", Traits)
	this:SetObjVar("RelConcerns", Concerns)
end

-- Loads all data from save
function Load()
	if initializer ~= nil then
		Traits = initializer.Traits or {}
		Concerns = initializer.Concerns or {}
		Save()
	end

	if this:HasObjVar("RelTraits") then
		Traits = this:GetObjVar("RelTraits")
	end if this:HasObjVar("RelConcerns") then
		Concerns = this:GetObjVar("RelConcerns")
	end
end

-- Flushes relation cache fully
function FlushCache()
	RelationCache = {}
end

-- Removes relation target given target from cache
function FlushSympathy(target)
	RelationCache[target] = nil
end

function CacheRelation(target, value)
	RelationCache[target] = {update = DateTime.UtcNow.Ticks, value = value}
end

-- Checks and sends sympathy to asker gameobj
function SendSympathy(callback, asker, target)
	local target = target or asker
	local sympathy = GetSympathy(target)
	asker:SendMessage("SympathyResponse", callback, this, target, sympathy)
end

function SendTrait(callback, asker, name)
	asker:SendMessage("TraitResponse", callback, this, name, Traits[name])
end

function SendConcern(callback, asker, name)
	asker:SendMessage("ConcernResponse", callback, this, name, Concerns[name])
end

SympathyRecievers = {}

function RecieveSympathy(callback, sender, target, sympathy)
	for name,func in pairs(SympathyRecievers) do
		func(callback, sender, target, sympathy)
	end
end

TraitRecievers = {}

function RecieveTrait(callback, sender, name, trait)
	for name,func in pairs(TraitRecievers) do
		func(callback, sender, name, trait)
	end
end

ConcernRecievers = {}

function RecieveConcern(callback, sender, name, concern)
	for name,func in pairs(ConcernRecievers) do
		func(callback, sender, name, concern)
	end
end

AllConditions.Sympathy = function (param, target)
	local sympathy = GetSympathy(target)
	local val
	if type(param) == "number" then
		val = param
	else
		val = param.value
	end

	if val ~= nil then return sympathy == val end

	local min = param.min or param[1] or -10000
	local max = param.max or param[2] or 10000

	return sympathy >= min and sympathy <= max
end

RegisterEventHandler(EventType.Message, "SetTrait", SetTrait)
RegisterEventHandler(EventType.Message, "ModifyTrait", ModifyTrait)
RegisterEventHandler(EventType.Message, "SetConcern", SetConcern)
RegisterEventHandler(EventType.Message, "SaveCharacter", Save)
RegisterEventHandler(EventType.Message, "FlushSympathyCache", FlushCache)
RegisterEventHandler(EventType.Message, "FlushSympathy", FlushSympathy)
RegisterEventHandler(EventType.Message, "GetSympathy", SendSympathy)
RegisterEventHandler(EventType.Message, "GetTrait", SendTrait)
RegisterEventHandler(EventType.Message, "GetConcern", SendConcern)

RegisterEventHandler(EventType.Message, "SympathyResponse", RecieveSympathy)
RegisterEventHandler(EventType.Message, "TraitResponse", RecieveTrait)
RegisterEventHandler(EventType.Message, "ConcernResponse", RecieveResponse)

Load()

-- Check for modular AI
if MODULAR_AI then
	Modules.Relations = true
end